import random
import numpy as np
from ..IMetaheuristic import IMetaheuristic
from ..ISolution import SolutionBasic
from typing import List, Callable

class BatAlgorithm(IMetaheuristic):
    def __init__(self, min_value: float, max_value: float, ndims: int, to_max: bool,
     objective_function: Callable[[List[float]], float],
      repair_function: Callable[[List[float]], List[float]],
      preprocess_function: Callable[[List[float]], List[float]] = None):
        """
        Initialize Bat Algorithm
        
        Args:
            to_max: Whether to maximize (True) or minimize (False)
            population_size: Number of bats
            fmin: Minimum frequency
            fmax: Maximum frequency
            loudness: Initial loudness
            pulse_rate: Initial pulse rate
            alpha: Loudness reduction constant
            gamma: Pulse rate increase constant
        """
        super().__init__(to_max)
        self._min = min_value
        self._max = max_value
        self._ndims = ndims
        self._to_max = to_max

        self.objective_function = objective_function
        self.preprocess_function = \
         preprocess_function if preprocess_function is not None else lambda p: p
        self.repair_function = \
         repair_function if repair_function is not None else lambda p: p

        self._bats = []
        self._velocities = []
        self._frequencies = []
        self._best_solution = None
        
    def initialize_population(self, population_size):
        """Initialize the bat population"""
        self._bats = []
        self._velocities = []
        self._frequencies = []
        
        for _ in range(population_size):
            # Initialize position
            point = [random.uniform(self._min, self._max) for _ in range(self._ndims)]
            point = self.cut_mod_point(point, self._min, self._max)
            repaired_point = self.repair_function(point)
            fitness = self.objective_function(self.preprocess_function(repaired_point))
            if fitness is not False and fitness is not None:
                self._bats.append(SolutionBasic(repaired_point, fitness))
                
                # Initialize velocity
                velocity = [0.0] * self._ndims
                self._velocities.append(velocity)
                
                # Initialize frequency
                self._frequencies.append(0.0)
        
        # Ensure we have at least one valid solution
        while len(self._bats) < 1:
            point = [random.uniform(self._min, self._max) for _ in range(self._ndims)]
            point = self.cut_mod_point(point, self._min, self._max)
            repaired_point = self.repair_function(point)
            fitness = self.objective_function(self.preprocess_function(repaired_point))
            if fitness is not False and fitness is not None:
                self._bats.append(SolutionBasic(repaired_point, fitness))
                velocity = [0.0] * self._ndims
                self._velocities.append(velocity)
                self._frequencies.append(0.0)
        
        self._best_solution = self.find_best_solution(self._bats)
    
    def _local_search(self, bat):
        """Perform local search around the bat"""
        epsilon = random.uniform(-1, 1)
        average_loudness = self._loudness
        new_position = []
        
        for i in range(self._ndims):
            new_pos = bat.point[i] + epsilon * average_loudness
            new_position.append(new_pos)
            
        return self.cut_mod_point(new_position, self._min, self._max)
    
    def run_yielded(self, iterations: int = 100, population: int = 30, fmin: float = 0, fmax: float = 2, A: float = 0.9, r0: float = 0.9, seed: int = None, verbose: bool = False):
        """Execute the Bat Algorithm with yielding"""
        self._iterations = iterations
        self._population = population
        self._seed = seed
        self._population_size = population
        self._fmin = fmin
        self._fmax = fmax
        self._loudness = A
        self._pulse_rate = r0
        self._alpha = 0.9
        self._gamma = 0.9
        
        random.seed(seed)
        np.random.seed(seed)

        self.initialize_population(self._population_size)
        
        iteration = 1
        best_solution_historical = self.find_best_solution(self._bats)
        best_fitness_historical = best_solution_historical.fitness
        best_point_historical = np.copy(best_solution_historical.point)
        
        # yield initial state
        points = [e.point for e in self._bats]
        fts = [e.fitness for e in self._bats]
        bin_point = self.preprocess_function(best_point_historical)
        yield iteration, best_fitness_historical, bin_point, points, fts
        
        while iteration <= iterations:
            if verbose:
                print("it: ", iteration, " fitness mejor: ", best_fitness_historical)
                
            for i in range(self._population_size):
                # Update frequency
                self._frequencies[i] = self._fmin + (self._fmax - self._fmin) * random.random()
                
                # Update velocity and position
                new_position = []
                for j in range(self._ndims):
                    # Update velocity
                    self._velocities[i][j] = (self._velocities[i][j] + 
                                            (self._bats[i].point[j] - self._best_solution.point[j]) * 
                                            self._frequencies[i])
                    
                    # Update position
                    new_pos = self._bats[i].point[j] + self._velocities[i][j]
                    new_position.append(new_pos)
                
                # Apply bounds
                new_position = self.cut_mod_point(new_position, self._min, self._max)
                
                # Local search with probability pulse_rate
                if random.random() > self._pulse_rate:
                    new_position = self._local_search(self._bats[i])
                
                # Evaluate new solution
                repaired_position = self.repair_function(new_position)
                new_fitness = self.objective_function(self.preprocess_function(repaired_position))
                
                # Accept new solution with probability loudness
                if (new_fitness is not False and new_fitness is not None and
                    random.random() < self._loudness and 
                    ((self._to_max and new_fitness > self._bats[i].fitness) or
                     (not self._to_max and new_fitness < self._bats[i].fitness))):
                    self._bats[i] = SolutionBasic(repaired_position, new_fitness)
                    
                    # Update pulse rate and loudness
                    self._pulse_rate = self._pulse_rate * (1 - np.exp(-self._gamma))
                    self._loudness *= self._alpha
            
            # Update best solution
            current_best = self.find_best_solution(self._bats)
            if ((self._to_max and current_best.fitness > best_fitness_historical) or
                (not self._to_max and current_best.fitness < best_fitness_historical)):
                best_fitness_historical = current_best.fitness
                best_point_historical = np.copy(current_best.point)
                self._best_solution = current_best
            
            iteration += 1
                
            # yield current state
            points = [e.point for e in self._bats]
            fts = [e.fitness for e in self._bats]
            bin_point = self.preprocess_function(best_point_historical)
            yield iteration, best_fitness_historical, bin_point, points, fts
        
        # Final yield
        bin_point = self.preprocess_function(best_point_historical)
        points = [e.point for e in self._bats]
        fts = [e.fitness for e in self._bats]
        yield iteration, best_fitness_historical, bin_point, points, fts

    def run(self, iterations: int = 100, population: int = 30, fmin: float = 0, fmax: float = 2, A: float = 0.9, r0: float = 0.9, seed: int = None, verbose: bool = False):
        """Execute the Bat Algorithm"""
        for _, best_fitness_historical, bin_point, _, _ in self.run_yielded(
            iterations, population, fmin, fmax, A, r0, seed, verbose):
            continue
        return best_fitness_historical, bin_point



import anmetal.utils.points_utils as utils
from anmetal.population.IMetaheuristic import IMetaheuristic
from anmetal.population.ISolution import SolutionWithId
import numpy as np
from numpy.random import RandomState
from typing import List, Callable

class AFSAMH_Real(IMetaheuristic):
    def __init__(self, min_value: float, max_value: float, ndims: int, to_max: bool,
     objective_function: Callable[[List[float]], float],
      repair_function: Callable[[List[float]], List[float]],
      preprocess_function: Callable[[List[float]], List[float]] = None):
        self._swarm = []
        self._random_generator : RandomState = None
        
        self._min = min_value
        self._max = max_value
        self._ndims = ndims
        self._to_max = to_max
        self.veces_movimiento = dict(move=0, prey=0, follow=0, swarm=0, leap=0)

        self.objective_function = objective_function
        self.preprocess_function = \
         preprocess_function if preprocess_function is not None else lambda p: p
        self.repair_function = \
         repair_function if repair_function is not None else lambda p: p


    
    def run_yielded(self, iterations: int = 100, population: int =30, verbose:bool=False,
     stagnation_variation: float =0.2, its_stagnation: int =5, leap_percentage: float =0.5, \
      velocity_percentage: float =0.3, n_points_to_choose: int =1, crowded_percentage: float =0.9,\
       visual_distance_percentage: float =0.1, seed: int = None):
        self._iterations = iterations
        self._population = population
        self._seed = seed
        self._visual_distance = np.sqrt(pow(self._max, 2) * self._ndims) * visual_distance_percentage
        self.n_points_to_choose = n_points_to_choose
        self._velocity_percentage = velocity_percentage
        self._random_generator = RandomState(seed)

        self.initialize_population(population)
        iteration = 1
        tau = 1
        best_solution_historical = self.find_best_solution(self._swarm)
        best_fitness_historical = best_solution_historical.fitness
        best_point_historical = np.copy(best_solution_historical.point)
        fitness_anterior_estancado = best_fitness_historical
        #yield
        points = [e.point for e in self._swarm]
        fts = [e.fitness for e in self._swarm]
        bin_point = self.preprocess_function(best_point_historical)
        yield iteration, best_fitness_historical, bin_point, points, fts
        while iteration <= iterations:
            if verbose:
                print("it: ", iteration, " fitness mejor: ", best_fitness_historical)
            for fish in self._swarm:
                neighborhood, neigh_points = self.get_neighborhood(fish)
                if len(neigh_points) == 0:
                    result_point = self.AF_Move(fish)
                else:
                    if len(neigh_points) / len(self._swarm) >= crowded_percentage:
                        result_point = self.AF_Follow(fish, neighborhood)
                    else:
                        result_point1 = self.AF_Swarm(fish, neighborhood)
                        result_point2 = self.AF_Prey(fish, neighborhood)
                        result_point = self.find_best_point([result_point1, result_point2])
                #result_point = self.borrar_1s_extra(result_point)
                result_point, fitness = self.repair_or_not(result_point)
                fish.move_to(result_point, fitness)# self.objective_function(self.preprocess_function(result_point)))
                # print("fitness del fish: ",fish.fish_id," es: ",fish.fitness)
            if iteration == tau * its_stagnation:
                fitness_mejor_actual = self.find_best_solution(self._swarm).fitness
                variation = 0
                if fitness_anterior_estancado != 0: #to avoid dividing by zero
                    variation = np.abs(fitness_anterior_estancado - fitness_mejor_actual) / fitness_anterior_estancado
                # print("variación: ", variation)
                if variation < stagnation_variation:
                    self.AF_Leap(leap_percentage)
                fitness_anterior_estancado = fitness_mejor_actual
                tau += 1
            iteration += 1
            # print("seteando historicoooooooooooooooooooooooo")
            best_solution_it = self.find_best_solution(self._swarm)
            best_fitness_it = best_solution_it.fitness
            best_point_it = np.copy(best_solution_it.point)
            if self._to_max and best_fitness_it > best_fitness_historical:
                best_fitness_historical = best_fitness_it
                best_point_historical = best_point_it
            if not self._to_max and best_fitness_it < best_fitness_historical:
                best_fitness_historical = best_fitness_it
                best_point_historical = best_point_it
            #yield
            points = [e.point for e in self._swarm]
            fts = [e.fitness for e in self._swarm]
            bin_point = self.preprocess_function(best_point_historical)
            yield iteration, best_fitness_historical, bin_point, points, fts
        # print("veces por movimiento: ", self.veces_movimiento)
        # print("terminó: ", mejor_fitness_historico)
        #yield
        points = [e.point for e in self._swarm]
        fts = [e.fitness for e in self._swarm]
        bin_point = self.preprocess_function(best_point_historical)
        yield iteration, best_fitness_historical, bin_point, points, fts

    def run(self, iterations: int = 100, population: int =30, verbose:bool=False,
     stagnation_variation: float =0.2, its_stagnation: int =5, leap_percentage: float =0.5, \
      velocity_percentage: float =0.3, n_points_to_choose: int =1, crowded_percentage: float =0.9,\
       visual_distance_percentage: float =0.1, seed: int = None):
        for _, best_fitness_historical, bin_point, _, _ in self.run_yielded(
            iterations, population, verbose, stagnation_variation, its_stagnation,
            leap_percentage, velocity_percentage, n_points_to_choose, crowded_percentage,
            visual_distance_percentage, seed):
            continue
        return best_fitness_historical, bin_point
    
    def initialize_population(self, population: int):
        for fishIndex in range(0, population):
            point, fitness = self.generate_random_point()
            fish = SolutionWithId(fishIndex, point, fitness)
            self._swarm.append(fish)



    def AF_Move(self, fish: SolutionWithId): #move randomly in visual distance
        self.veces_movimiento["move"] += 1
        points = []
        for _ in range(self.n_points_to_choose):
            point = self.generate_in_visual(fish.point)
            points.append(point)
        return self.find_best_point(points)

    def AF_Swarm(self, fish: SolutionWithId, neighborhood: List[SolutionWithId]): #move to center of neighborhood
        self.veces_movimiento["swarm"] += 1
        central_point = self.get_central_point(neighborhood)
        if self.is_point_valid(central_point) \
         and np.array_equal(self.find_best_point([central_point, fish.point]), central_point):
            return central_point
        else:
            return self.AF_Follow(fish, neighborhood)

    def AF_Prey(self, fish: SolutionWithId, neighborhood: List[SolutionWithId]): #move to best neighbour
        self.veces_movimiento["prey"] +=1
        # neigh_fishes, neigh_points = self.get_neighborhood(fish)
        best_neighbour = self.find_best_solution(neighborhood)
        if self.find_best_solution([best_neighbour, fish]) == fish: # == best_neighbour: #the original is to best_neigh
            return self.get_closer_to(fish.point, best_neighbour.point)
        else:
            return self.AF_Follow(fish, neighborhood)

    def AF_Follow(self, fish: SolutionWithId, neighborhood: List[SolutionWithId]): #move to any neighbour
        self.veces_movimiento["follow"] += 1
        # neighFishes, neighPoints = self.get_neighborhood(fish)
        points = []
        for _ in range(self.n_points_to_choose):
            neighbor_to = int(self._random_generator.uniform(0, len(neighborhood) - 1))
            points.append(neighborhood[neighbor_to].point)
        return self.get_closer_to(fish.point, self.find_best_point(points))

    def AF_Leap(self, leap_percentage: float): #escape
        self.veces_movimiento["leap"] += 1
        num_fishes_to_leap = int(self._population * leap_percentage)
        index_fishes_to_leap = [i for i in range(self._population)]
        self._random_generator.shuffle(index_fishes_to_leap)
        index_fishes_to_leap = index_fishes_to_leap[0:num_fishes_to_leap]
        for fish_index in index_fishes_to_leap:
            # fish_index = int(np.floor(config.random_generator.uniform(0, self.config["fishNumber"])))
            gen_point, gen_fitness = self.generate_random_point()
            self._swarm[fish_index].move_to(gen_point, gen_fitness)


    def get_neighborhood(self, fish: SolutionWithId):
        neighborhood = []
        points_neighborhood = []
        for f in self._swarm:
            if f._id != fish._id:
                distance = utils.distance(fish.point, f.point)
                if distance <= self._visual_distance:
                    neighborhood.append(f)
                    points_neighborhood.append(f.point)
        return neighborhood, points_neighborhood

    def get_central_point(self, neighborhood: List[SolutionWithId]):
        media_dim = []
        for i in range(0, self._ndims):
            media_dim.append(0)
        # neigh_fishes, neigh_points = self.get_neighborhood(fish)
        for neighbor in neighborhood:
            p = neighbor.point
            for i in range(0, len(media_dim)):
                media_dim[i] += p[i]
        for i in range(0, len(media_dim)):
            media_dim[i] /= len(neighborhood)
        return self.repair_or_not(media_dim)[0]

    def is_in_visual(self, fish: SolutionWithId, point: List[float]):
        distance = utils.distance(fish.point, point)
        if distance <= self._visual_distance:
            return True
        else:
            return False

    def generate_in_visual(self, origin_point: List[float]):
        distance = self._random_generator.uniform(0 , self._visual_distance, 1)
        angles = self._random_generator.uniform(0 , 2*np.pi, self._ndims -1)
        cartesian_point = utils.nsphere_to_cartesian(distance, angles)
        for i in range(0, len(cartesian_point)):
            cartesian_point[i] += origin_point[i] #scale
        return self.repair_or_not(cartesian_point)[0]

    def generate_random_point(self):
        cartesian_point = self._random_generator.uniform(self._min,
                                                        self._max,
                                                        self._ndims)
        return self.repair_or_not(cartesian_point)
    
    def repair_or_not(self, cartesian_point: List[float]):
        cartesian_point = IMetaheuristic.cut_mod_point(cartesian_point, self._min, self._max)
        fitness = self.objective_function(self.preprocess_function(cartesian_point))
        if not fitness:
            new_point = self.repair_function(cartesian_point)
            fitness = self.objective_function(self.preprocess_function(new_point))
            return new_point, fitness
        else:
            return cartesian_point, fitness
    
    def get_closer_to(self, origin_point: List[float], dest_point: List[float]):
        for idim in range(len(origin_point)):
            origin_point[idim] += self._velocity_percentage * (origin_point[idim] - dest_point[idim])
        return self.repair_or_not(origin_point)[0]


import random
import numpy as np
from ..IMetaheuristic import IMetaheuristic
from ..ISolution import SolutionBasic

class FireflyAlgorithm(IMetaheuristic):
    def __init__(self, min_value, max_value, ndims, to_max, objective_function, repair_function, preprocess_function=None):
        """
        Initialize Firefly Algorithm
        
        Args:
            min_value: Minimum value for variables
            max_value: Maximum value for variables
            ndims: Number of dimensions
            to_max: Whether to maximize (True) or minimize (False)
            objective_function: Function to optimize
            repair_function: Function to repair invalid solutions
            preprocess_function: Function to preprocess solutions
        """
        super().__init__(to_max)
        self._min = min_value
        self._max = max_value
        self._ndims = ndims
        self._to_max = to_max
        self._objective_function = objective_function
        self._repair_function = repair_function if repair_function is not None else lambda p: p
        self._preprocess_function = preprocess_function if preprocess_function is not None else lambda p: p
        # Default parameters - will be overridden in run methods
        self._population_size = 30
        self._alpha = 0.5
        self._beta0 = 1.0
        self._gamma = 1.0
        self._fireflies = []
        self._best_solution = None
    
    def initialize_population(self):
        """Initialize the population of fireflies"""
        self._fireflies = []
        for _ in range(self._population_size):
            point = [random.uniform(self._min, self._max) for _ in range(self._ndims)]
            if self._repair_function:
                point = self._repair_function(point)
                
            fitness = self._objective_function(self._preprocess_function(point))
            if fitness is not False and fitness is not None:
                self._fireflies.append(SolutionBasic(point, fitness))
        
        if self._fireflies:
            self._best_solution = self.find_best_solution(self._fireflies)
    
    def _distance(self, firefly1, firefly2):
        """Calculate the Cartesian distance between two fireflies"""
        return np.sqrt(sum((x1 - x2) ** 2 for x1, x2 in zip(firefly1.point, firefly2.point)))
    
    def _attract_and_move(self, firefly1, firefly2):
        """Move firefly1 towards firefly2 if firefly2 is brighter"""
        distance = self._distance(firefly1, firefly2)
        beta = self._beta0 * np.exp(-self._gamma * distance ** 2)
        
        new_position = []
        for i in range(self._ndims):
            # Movement towards brighter firefly + random movement
            rand = random.uniform(-0.5, 0.5)
            movement = (beta * (firefly2.point[i] - firefly1.point[i]) + 
                       self._alpha * rand)
            new_pos = firefly1.point[i] + movement
            new_position.append(new_pos)
        
        # Apply boundary conditions
        return [max(self._min, min(self._max, pos)) for pos in new_position]
    
    def run_yielded(self, iterations: int = 100, population: int = 30, alpha: float = 0.5, beta0: float = 1.0, gamma: float = 1.0, seed: int = None, verbose: bool = False):
        """Execute the Firefly Algorithm with yielding"""
        self._iterations = iterations
        self._population_size = population  # Update population size
        self._alpha = alpha  # Update alpha parameter
        self._beta0 = beta0  # Update beta0 parameter
        self._gamma = gamma  # Update gamma parameter
        self._seed = seed
        
        random.seed(seed)
        np.random.seed(seed)
        
        self.initialize_population()
        
        if not self._fireflies or not self._best_solution:
            return
            
        iteration = 1
        best_solution_historical = self._best_solution
        best_fitness_historical = best_solution_historical.fitness
        best_point_historical = np.copy(best_solution_historical.point)
        
        # yield initial state
        points = [e.point for e in self._fireflies]
        fts = [e.fitness for e in self._fireflies]
        bin_point = self._preprocess_function(best_point_historical)
        yield iteration, best_fitness_historical, bin_point, points, fts
            
        for it in range(iterations):
            if verbose:
                print("it: ", it+1, " fitness mejor: ", best_fitness_historical)
                
            # For each firefly
            for i in range(len(self._fireflies)):
                # Compare with all other fireflies
                for j in range(len(self._fireflies)):
                    if i == j:
                        continue
                        
                    # Move if the other firefly is brighter
                    if ((self._to_max and self._fireflies[j].fitness > self._fireflies[i].fitness) or
                        (not self._to_max and self._fireflies[j].fitness < self._fireflies[i].fitness)):
                        new_position = self._attract_and_move(self._fireflies[i], self._fireflies[j])
                        
                        if self._repair_function:
                            new_position = self._repair_function(new_position)
                            
                        new_fitness = self._objective_function(self._preprocess_function(new_position))
                        if new_fitness is not False and new_fitness is not None:
                            self._fireflies[i] = SolutionBasic(new_position, new_fitness)
            
            # Update best solution
            if self._fireflies:
                current_best = self.find_best_solution(self._fireflies)
                if current_best and self._best_solution:
                    if ((self._to_max and current_best.fitness > self._best_solution.fitness) or
                        (not self._to_max and current_best.fitness < self._best_solution.fitness)):
                        self._best_solution = current_best
            
            # Reduce alpha (optional)
            self._alpha *= 0.97
            
            iteration += 1
            
            # Update best solution
            if ((self._to_max and self._best_solution.fitness > best_fitness_historical) or
                (not self._to_max and self._best_solution.fitness < best_fitness_historical)):
                best_fitness_historical = self._best_solution.fitness
                best_point_historical = np.copy(self._best_solution.point)
                
            # yield current state
            points = [e.point for e in self._fireflies]
            fts = [e.fitness for e in self._fireflies]
            bin_point = self._preprocess_function(best_point_historical)
            yield iteration, best_fitness_historical, bin_point, points, fts
    
    def run(self, iterations: int = 100, population: int = 30, alpha: float = 0.5, beta0: float = 1.0, gamma: float = 1.0, seed: int = None, verbose: bool = False):
        """Execute the Firefly Algorithm"""
        for _, best_fitness_historical, bin_point, _, _ in self.run_yielded(
            iterations, population, alpha, beta0, gamma, seed, verbose):
            continue
        return best_fitness_historical, bin_point



#This file test the Real version of the Metaheuristics (Numerical) with the problems in anmetal.problems.nphard_real package
from anmetal.population.AFSA.AFSAMH_Real import AFSAMH_Real
from anmetal.population.SillyRandom.GreedyMH_Real import GreedyMH_Real
from anmetal.population.SillyRandom.GreedyMH_Real_WithLeap import GreedyMH_Real_WithLeap
from anmetal.population.PSO.PSOMH_Real import PSOMH_Real
from anmetal.population.PSO.PSOMH_Real_WithLeap import PSOMH_Real_WithLeap
from anmetal.population.ABC.abc import ArtificialBeeColony as ABCMH_Real
from anmetal.population.AntColony.aco import AntColony as ACOMH_Real
from anmetal.population.Bat.bat import BatAlgorithm as BatMH_Real
from anmetal.population.Blackhole.blackhole import BlackHole as BlackholeMH_Real
from anmetal.population.Cuckoo.cuckoo import CuckooSearch as CuckooMH_Real
from anmetal.population.Firefly.firefly import FireflyAlgorithm as FireflyMH_Real
from anmetal.population.HarmonySearch.harmony import HarmonySearch as HSMH_Real

from anmetal.problems.nphard_real.partition__and_subset_sum import Partition_Real, Subset_Real

to_use = [
    "AFSA",
    "Greedy",
    "GreedyWL",
    "PSO",
    "PSOWL",
    "ABC",
    "ACO",
    "BAT",
    "BH",
    "CUCKOO",
    "FIREFLY",
    "HS"
    ]

partition_problem = Partition_Real(seed=0, num_dims=200)
subset_problem = Subset_Real(seed=0, num_dims=200)


problem_to_solve = "partition sum"
#problem_to_solve = "subset sum"

to_verbose = True

if "AFSA" in to_use:
    print("create afsa")
    if problem_to_solve == "partition sum":
        mh = AFSAMH_Real(partition_problem.min_x, partition_problem.max_x, partition_problem.ndim, False, partition_problem.objective_function, partition_problem.repair_function, partition_problem.preprocess_function)
    if problem_to_solve == "subset sum":
        mh = AFSAMH_Real(subset_problem.min_x, subset_problem.max_x, subset_problem.ndim, False, subset_problem.objective_function, subset_problem.repair_function, subset_problem.preprocess_function)
    print("to run afsa")
    fit, pt = mh.run(verbose=to_verbose, visual_distance_percentage=0.5, velocity_percentage=0.5, n_points_to_choose=3, crowded_percentage=0.7, its_stagnation=4, leap_percentage=0.3, stagnation_variation=0.4, seed=115)
    print(fit)
    print(pt)
    if problem_to_solve == "partition sum":
        partition_problem.print_difference_subsets(pt)
    if problem_to_solve == "subset sum":
        subset_problem.print_difference_subsets(pt)

if "GreedyWL" in to_use:
    print("create GreedyWL")
    if problem_to_solve == "partition sum":
        mh = GreedyMH_Real_WithLeap(partition_problem.min_x, partition_problem.max_x, partition_problem.ndim, False, partition_problem.objective_function, partition_problem.repair_function, partition_problem.preprocess_function)
    if problem_to_solve == "subset sum":
        mh = GreedyMH_Real_WithLeap(subset_problem.min_x, subset_problem.max_x, subset_problem.ndim, False, subset_problem.objective_function, subset_problem.repair_function, subset_problem.preprocess_function)
    print("to run greedy")
    fit, pt = mh.run(verbose=to_verbose, iterations=100, population=30, stagnation_variation=0.4, its_stagnation=5, leap_percentage=0.8, seed=115)
    print(fit)
    print(pt)
    if problem_to_solve == "partition sum":
        partition_problem.print_difference_subsets(pt)
    if problem_to_solve == "subset sum":
        subset_problem.print_difference_subsets(pt)

if "Greedy" in to_use:
    print("create Greedy")
    if problem_to_solve == "partition sum":
        mh = GreedyMH_Real(partition_problem.min_x, partition_problem.max_x, partition_problem.ndim, False, partition_problem.objective_function, partition_problem.repair_function, partition_problem.preprocess_function)
    if problem_to_solve == "subset sum":
        mh = GreedyMH_Real(subset_problem.min_x, subset_problem.max_x, subset_problem.ndim, False, subset_problem.objective_function, subset_problem.repair_function, subset_problem.preprocess_function)
    print("to run greedy")
    fit, pt = mh.run(verbose=to_verbose, iterations=100, population=30, seed=115)
    print(fit)
    print(pt)
    if problem_to_solve == "partition sum":
        partition_problem.print_difference_subsets(pt)
    if problem_to_solve == "subset sum":
        subset_problem.print_difference_subsets(pt)

if "PSO" in to_use:
    print("create PSO")
    if problem_to_solve == "partition sum":
        mh = PSOMH_Real(partition_problem.min_x, partition_problem.max_x, partition_problem.ndim, False, partition_problem.objective_function, partition_problem.repair_function, partition_problem.preprocess_function)
    if problem_to_solve == "subset sum":
        mh = PSOMH_Real(subset_problem.min_x, subset_problem.max_x, subset_problem.ndim, False, subset_problem.objective_function, subset_problem.repair_function, subset_problem.preprocess_function)
    print("to run PSO")
    fit, pt = mh.run(verbose=to_verbose, iterations=100, population=30, omega=0.8, phi_g=1, phi_p=0.5,seed=115)
    print(fit)
    print(pt)
    if problem_to_solve == "partition sum":
        partition_problem.print_difference_subsets(pt)
    if problem_to_solve == "subset sum":
        subset_problem.print_difference_subsets(pt)

if "PSOWL" in to_use:
    print("create PSOWL")
    if problem_to_solve == "partition sum":
        mh = PSOMH_Real_WithLeap(partition_problem.min_x, partition_problem.max_x, partition_problem.ndim, False, partition_problem.objective_function, partition_problem.repair_function, partition_problem.preprocess_function)
    if problem_to_solve == "subset sum":
        mh = PSOMH_Real_WithLeap(subset_problem.min_x, subset_problem.max_x, subset_problem.ndim, False, subset_problem.objective_function, subset_problem.repair_function, subset_problem.preprocess_function)
    print("to run PSO")
    fit, pt = mh.run(verbose=to_verbose, iterations=100, population=30, omega=0.8, phi_g=1, phi_p=0.5 ,seed=115, stagnation_variation=0.4, its_stagnation=5, leap_percentage=0.8)
    print(fit)
    print(pt)
    if problem_to_solve == "partition sum":
        partition_problem.print_difference_subsets(pt)
    if problem_to_solve == "subset sum":
        subset_problem.print_difference_subsets(pt)

if "ABC" in to_use:
    print("create ABC")
    if problem_to_solve == "partition sum":
        mh = ABCMH_Real(partition_problem.min_x, partition_problem.max_x, partition_problem.ndim, False, partition_problem.objective_function, partition_problem.repair_function, partition_problem.preprocess_function)
    if problem_to_solve == "subset sum":
        mh = ABCMH_Real(subset_problem.min_x, subset_problem.max_x, subset_problem.ndim, False, subset_problem.objective_function, subset_problem.repair_function, subset_problem.preprocess_function)
    print("to run ABC")
    fit, pt = mh.run(verbose=to_verbose, iterations=100, population=30, limit=20, seed=115)
    print(fit)
    print(pt)
    if problem_to_solve == "partition sum":
        partition_problem.print_difference_subsets(pt)
    if problem_to_solve == "subset sum":
        subset_problem.print_difference_subsets(pt)

if "ACO" in to_use:
    print("create ACO")
    if problem_to_solve == "partition sum":
        mh = ACOMH_Real(partition_problem.min_x, partition_problem.max_x, partition_problem.ndim, False, partition_problem.objective_function, partition_problem.repair_function, partition_problem.preprocess_function)
    if problem_to_solve == "subset sum":
        mh = ACOMH_Real(subset_problem.min_x, subset_problem.max_x, subset_problem.ndim, False, subset_problem.objective_function, subset_problem.repair_function, subset_problem.preprocess_function)
    print("to run ACO")
    fit, pt = mh.run(verbose=to_verbose, iterations=100, population=30, evaporation_rate=0.1, alpha=1.0, beta=2.0, seed=115)
    print(fit)
    print(pt)
    if problem_to_solve == "partition sum":
        partition_problem.print_difference_subsets(pt)
    if problem_to_solve == "subset sum":
        subset_problem.print_difference_subsets(pt)

if "BAT" in to_use:
    print("create BAT")
    if problem_to_solve == "partition sum":
        mh = BatMH_Real(partition_problem.min_x, partition_problem.max_x, partition_problem.ndim, False, partition_problem.objective_function, partition_problem.repair_function, partition_problem.preprocess_function)
    if problem_to_solve == "subset sum":
        mh = BatMH_Real(subset_problem.min_x, subset_problem.max_x, subset_problem.ndim, False, subset_problem.objective_function, subset_problem.repair_function, subset_problem.preprocess_function)
    print("to run BAT")
    fit, pt = mh.run(verbose=to_verbose, iterations=100, population=30, fmin=0, fmax=2, A=0.9, r0=0.9, seed=115)
    print(fit)
    print(pt)
    if problem_to_solve == "partition sum":
        partition_problem.print_difference_subsets(pt)
    if problem_to_solve == "subset sum":
        subset_problem.print_difference_subsets(pt)

if "BH" in to_use:
    print("create Blackhole")
    if problem_to_solve == "partition sum":
        mh = BlackholeMH_Real(partition_problem.min_x, partition_problem.max_x, partition_problem.ndim, False, partition_problem.objective_function, partition_problem.repair_function, partition_problem.preprocess_function)
    if problem_to_solve == "subset sum":
        mh = BlackholeMH_Real(subset_problem.min_x, subset_problem.max_x, subset_problem.ndim, False, subset_problem.objective_function, subset_problem.repair_function, subset_problem.preprocess_function)
    print("to run Blackhole")
    fit, pt = mh.run(verbose=to_verbose, iterations=100, population=30, seed=115)
    print(fit)
    print(pt)
    if problem_to_solve == "partition sum":
        partition_problem.print_difference_subsets(pt)
    if problem_to_solve == "subset sum":
        subset_problem.print_difference_subsets(pt)

if "CUCKOO" in to_use:
    print("create Cuckoo")
    if problem_to_solve == "partition sum":
        mh = CuckooMH_Real(partition_problem.min_x, partition_problem.max_x, partition_problem.ndim, False, partition_problem.objective_function, partition_problem.repair_function, partition_problem.preprocess_function)
    if problem_to_solve == "subset sum":
        mh = CuckooMH_Real(subset_problem.min_x, subset_problem.max_x, subset_problem.ndim, False, subset_problem.objective_function, subset_problem.repair_function, subset_problem.preprocess_function)
    print("to run Cuckoo")
    fit, pt = mh.run(verbose=to_verbose, iterations=100, population=30, pa=0.25, seed=115)
    print(fit)
    print(pt)
    if problem_to_solve == "partition sum":
        partition_problem.print_difference_subsets(pt)
    if problem_to_solve == "subset sum":
        subset_problem.print_difference_subsets(pt)

if "FIREFLY" in to_use:
    print("create Firefly")
    if problem_to_solve == "partition sum":
        mh = FireflyMH_Real(partition_problem.min_x, partition_problem.max_x, partition_problem.ndim, False, partition_problem.objective_function, partition_problem.repair_function, partition_problem.preprocess_function)
    if problem_to_solve == "subset sum":
        mh = FireflyMH_Real(subset_problem.min_x, subset_problem.max_x, subset_problem.ndim, False, subset_problem.objective_function, subset_problem.repair_function, subset_problem.preprocess_function)
    print("to run Firefly")
    fit, pt = mh.run(verbose=to_verbose, iterations=100, population=30, alpha=0.5, beta0=1.0, gamma=1.0, seed=115)
    print(fit)
    print(pt)
    if problem_to_solve == "partition sum":
        partition_problem.print_difference_subsets(pt)
    if problem_to_solve == "subset sum":
        subset_problem.print_difference_subsets(pt)

if "HS" in to_use:
    print("create Harmony Search")
    if problem_to_solve == "partition sum":
        mh = HSMH_Real(partition_problem.min_x, partition_problem.max_x, partition_problem.ndim, False, partition_problem.objective_function, partition_problem.repair_function, partition_problem.preprocess_function)
    if problem_to_solve == "subset sum":
        mh = HSMH_Real(subset_problem.min_x, subset_problem.max_x, subset_problem.ndim, False, subset_problem.objective_function, subset_problem.repair_function, subset_problem.preprocess_function)
    print("to run Harmony Search")
    fit, pt = mh.run(verbose=to_verbose, iterations=100, population=30, hmcr=0.9, par=0.3, bw=0.2, seed=115)
    print(fit)
    print(pt)
    if problem_to_solve == "partition sum":
        partition_problem.print_difference_subsets(pt)
    if problem_to_solve == "subset sum":
        subset_problem.print_difference_subsets(pt)



import random
import numpy as np
from ..IMetaheuristic import IMetaheuristic
from ..ISolution import SolutionBasic

class BlackHole(IMetaheuristic):
    def __init__(self, min_value, max_value, ndims, to_max=True, objective_function=None, repair_function=None, preprocess_function=None, population_size=30):
        super().__init__(to_max)
        self._min = min_value
        self._max = max_value
        self._ndims = ndims
        self._objective_function = objective_function
        self._repair_function = repair_function if repair_function is not None else lambda p: p
        self._preprocess_function = preprocess_function if preprocess_function is not None else lambda p: p
        self._population_size = population_size
        self._black_hole = None
        self._stars = []
        
    def initialize_population(self):
        """Initialize the population of stars randomly"""
        self._stars = []
        for _ in range(self._population_size):
            point = [random.uniform(self._min, self._max) for _ in range(self._ndims)]
            if self._repair_function:
                point = self._repair_function(point)
            if self._preprocess_function:
                point = self._preprocess_function(point)
            
            fitness = self._objective_function(point)
            if fitness is not False and fitness is not None:
                self._stars.append(SolutionBasic(point, fitness))
        
        # Find the initial black hole (best solution)
        if self._stars:
            self._black_hole = self.find_best_solution(self._stars)
        
    def run_yielded(self, iterations: int = 100, population: int = 30, seed: int = None, verbose: bool = False):
        """Execute the Black Hole algorithm with yielding"""
        self._iterations = iterations
        self._population = population
        self._seed = seed
        
        random.seed(seed)
        np.random.seed(seed)
        
        self.initialize_population()
        
        if not self._stars or not self._black_hole:
            return
            
        iteration = 1
        best_solution_historical = self._black_hole
        best_fitness_historical = best_solution_historical.fitness
        best_point_historical = np.copy(best_solution_historical.point)
        
        # yield initial state
        points = [e.point for e in self._stars]
        fts = [e.fitness for e in self._stars]
        bin_point = self._preprocess_function(best_point_historical)
        yield iteration, best_fitness_historical, bin_point, points, fts
            
        for it in range(iterations):
            if verbose:
                print("it: ", it+1, " fitness mejor: ", best_fitness_historical)
                
            # Calculate the radius of event horizon
            event_horizon = self._calculate_event_horizon()
            
            # Move stars towards black hole and check for absorption
            new_stars = []
            for star in self._stars:
                if star == self._black_hole:
                    continue
                    
                # Move star towards black hole
                new_position = self._move_star_towards_black_hole(star.point)
                
                # Check if star crosses event horizon
                if self._is_within_event_horizon(new_position, event_horizon):
                    # Generate new random star
                    new_point = [random.uniform(self._min, self._max) for _ in range(self._ndims)]
                    if self._repair_function:
                        new_point = self._repair_function(new_point)
                    if self._preprocess_function:
                        new_point = self._preprocess_function(new_point)
                        
                    fitness = self._objective_function(new_point)
                    if fitness is not False and fitness is not None:
                        new_stars.append(SolutionBasic(new_point, fitness))
                else:
                    # Update star position
                    if self._repair_function:
                        new_position = self._repair_function(new_position)
                    if self._preprocess_function:
                        new_position = self._preprocess_function(new_position)
                        
                    fitness = self._objective_function(new_position)
                    if fitness is not False and fitness is not None:
                        star.move_to(new_position, fitness)
                        new_stars.append(star)
                    else:
                        new_stars.append(star)
            
            self._stars = new_stars
            
            # Update black hole if better solution found
            if self._stars:
                current_best = self.find_best_solution(self._stars)
                if current_best and self._black_hole:
                    if ((self._to_max and current_best.fitness > self._black_hole.fitness) or
                        (not self._to_max and current_best.fitness < self._black_hole.fitness)):
                        self._black_hole = current_best
                        
            iteration += 1
            
            # Update best solution
            if ((self._to_max and self._black_hole.fitness > best_fitness_historical) or
                (not self._to_max and self._black_hole.fitness < best_fitness_historical)):
                best_fitness_historical = self._black_hole.fitness
                best_point_historical = np.copy(self._black_hole.point)
                
            # yield current state
            points = [e.point for e in self._stars]
            fts = [e.fitness for e in self._stars]
            bin_point = self._preprocess_function(best_point_historical)
            yield iteration, best_fitness_historical, bin_point, points, fts
    
    def run(self, iterations: int = 100, population: int = 30, seed: int = None, verbose: bool = False):
        """Execute the Black Hole algorithm"""
        for _, best_fitness_historical, bin_point, _, _ in self.run_yielded(
            iterations, population, seed, verbose):
            continue
        return best_fitness_historical, bin_point
    
    def _calculate_event_horizon(self):
        """Calculate the radius of the event horizon"""
        black_hole_fitness = abs(self._black_hole.fitness)
        total_fitness = sum(abs(star.fitness) for star in self._stars)
        return black_hole_fitness / total_fitness
    
    def _move_star_towards_black_hole(self, star_position):
        """Move a star towards the black hole"""
        new_position = []
        for i in range(len(star_position)):
            # Random number between 0 and 1
            r = random.random()
            # Move star towards black hole
            pos = star_position[i] + r * (self._black_hole.point[i] - star_position[i])
            new_position.append(pos)
        
        # Apply boundary conditions
        return [max(self._min, min(self._max, pos)) for pos in new_position]
    
    def _is_within_event_horizon(self, star_position, event_horizon):
        """Check if a star is within the event horizon of the black hole"""
        distance = np.sqrt(sum((s - b) ** 2 for s, b in zip(star_position, self._black_hole.point)))
        return distance < event_horizon



"""Tests for optimization problems."""
import pytest
import numpy as np
from anmetal.problems.nphard_categorical.knapsack import Knapsack_Categorical
from anmetal.problems.nphard_categorical.sudoku import Sudoku

class TestKnapsack:
    def test_knapsack_initialization(self):
        """Test knapsack problem initialization."""
        knapsack = Knapsack_Categorical(
            knapsack_capacity=50.8,
            total_posible_elements=50,
            seed=0,
            max_value=5,
            max_cost=6
        )
        assert knapsack.capacity == 50.8
        assert knapsack.total_posible_elements == 50

    def test_knapsack_objective_function(self):
        """Test knapsack objective function calculation."""
        knapsack = Knapsack_Categorical(
            knapsack_capacity=50.8,
            total_posible_elements=5,
            seed=0,
            max_value=5,
            max_cost=6
        )
        solution = [1, 0, 1, 0, 1]  # Example solution
        value = knapsack.objective_function(solution)
        assert isinstance(value, (int, float))
        assert value >= 0

class TestSudoku:
    def test_sudoku_initialization(self):
        """Test sudoku problem initialization."""
        initial_state = [[0] * 9 for _ in range(9)]
        sudoku = Sudoku(initial_state)
        assert len(sudoku.state) == 9
        assert all(len(row) == 9 for row in sudoku.state)

    def test_sudoku_is_valid(self):
        """Test sudoku validity checking."""
        initial_state = [[0] * 9 for _ in range(9)]
        sudoku = Sudoku(initial_state)
        # Test empty board validity
        assert sudoku.is_valid_solution(initial_state)



#functions taken from https://doi.org/10.1007/s00521-017-3088-3
#Applying modified cuckoo search algorithm for solving systems of nonlinear equations
import numpy as np

class Camelback:
    @staticmethod
    def func_simple(x: float, y:float) -> float:
        result = (4 - 2.1*(x**2) + (x**4)/3.0) * (x**2)
        result += x*y
        result += (4*(y**2) - 4) * (y**2)
        return result

    @staticmethod
    def func(x_array: list) -> float:
        return Camelback.func_simple(x_array[0], x_array[1])

    @staticmethod
    def get_limits():
        return [-3, 3], [-2, 2]
    
    @staticmethod
    def get_theoretical_optimum():
        return -1.03163
    
    @staticmethod
    def get_type():
        return "min"
    
class Goldsteinprice:
    @staticmethod
    def func_simple(x: float, y:float) -> float:
        result = 1 + ((x+y+1)**2) * (19-14*x + 3*(x**2)-14*y + 6*x*y + 3*(y**2))
        result = result * ((30+ (2*x - 3*y)**2) * 18-32*x+ 12*(x**2) + 48*y - 36*x*y + 27*(y**2) )
        return result

    @staticmethod
    def func(x_array: list) -> float:
        return Goldsteinprice.func_simple(x_array[0], x_array[1])
        
    @staticmethod
    def get_limits():
        return [-2, 2]
    
    @staticmethod
    def get_theoretical_optimum():
        return 3
    
    @staticmethod
    def get_type():
        return "min"

class Pshubert1:
    @staticmethod
    def func_simple(x: float, y:float) -> float:
        res1 = 0
        for i in range(1, 6):#1, 2, 3, 4, 5
            res1 += float(i)*np.cos( (i+1.0)*x + i )
        res2 = 0
        for i in range(1, 6):#1,2,3,4,5
            res2 += float(i)*np.cos( (i+1.0)*y + i )
        result = (0.5*( (x+1.42513)**2 + (y+0.80032)**2 ) )
        result = (res1*res2) + result
        return result

    @staticmethod
    def func(x_array: list) -> float:
        return Pshubert1.func_simple(x_array[0], x_array[1])
        
    @staticmethod
    def get_limits():
        return [-10, 10]
    
    @staticmethod
    def get_theoretical_optimum():
        return -186.73091
    
    @staticmethod
    def get_type():
        return "min"

class Pshubert2:
    @staticmethod
    def func_simple(x: float, y:float) -> float:
        res1 = 0
        for i in range(1, 6):#1, 2, 3, 4, 5
            res1 += float(i)*np.cos( (i+1.0)*x + i )
        res2 = 0
        for i in range(1, 6):#1,2,3,4,5
            res2 += float(i)*np.cos( (i+1.0)*y + i )
        result = (x+1.42513)**2 + (y+0.80032)**2
        result = (res1*res2) + result
        return result

    @staticmethod
    def func(x_array: list) -> float:
        return Pshubert2.func_simple(x_array[0], x_array[1])
        
    @staticmethod
    def get_limits():
        return [-10, 10]
    
    @staticmethod
    def get_theoretical_optimum():
        return -186.73091
    
    @staticmethod
    def get_type():
        return "min"
        
class Shubert:
    @staticmethod
    def func_simple(x: float, y:float) -> float:
        res1 = 0
        for i in range(1, 6):#1, 2, 3, 4, 5
            res1 += float(i)*np.cos( (i+1.0)*x + i )
        res2 = 0
        for i in range(1, 6):#1,2,3,4,5
            res2 += float(i)*np.cos( (i+1.0)*y + i )
        return res1*res2

    @staticmethod
    def func(x_array: list) -> float:
        return Shubert.func_simple(x_array[0], x_array[1])
        
    @staticmethod
    def get_limits():
        return [-10, 10]
    
    @staticmethod
    def get_theoretical_optimum():
        return -186.73091
    
    @staticmethod
    def get_type():
        return "min"
        
class Quartic:
    @staticmethod
    def func_simple(x: float, y:float) -> float:
        return (x**4/float(4)) - (x**2/float(2)) + (x/float(10)) + (y**2/float(2))

    @staticmethod
    def func(x_array: list) -> float:
        return Quartic.func_simple(x_array[0], x_array[1])
        
    @staticmethod
    def get_limits():
        return [-10, 10]
    
    @staticmethod
    def get_theoretical_optimum():
        return -0.35239
    
    @staticmethod
    def get_type():
        return "min"


import random
import numpy as np
from ..IMetaheuristic import IMetaheuristic
from ..ISolution import SolutionBasic

class HarmonySearch(IMetaheuristic):
    def __init__(self, min_value, max_value, ndims, to_max, objective_function, repair_function, preprocess_function=None):
        """
        Initialize Harmony Search Algorithm
        
        Args:
            min_value: Minimum value for variables
            max_value: Maximum value for variables
            ndims: Number of dimensions
            to_max: Whether to maximize (True) or minimize (False)
            objective_function: Function to optimize
            repair_function: Function to repair invalid solutions
            preprocess_function: Function to preprocess solutions
        """
        super().__init__(to_max)
        self._min = min_value
        self._max = max_value
        self._ndims = ndims
        self._to_max = to_max
        self._objective_function = objective_function
        self._repair_function = repair_function if repair_function is not None else lambda p: p
        self._preprocess_function = preprocess_function if preprocess_function is not None else lambda p: p
        # Default parameters - will be overridden in run methods
        self._hms = 30  # harmony memory size
        self._hmcr = 0.9  # harmony memory considering rate
        self._par = 0.3  # pitch adjustment rate
        self._bw = 0.01  # bandwidth
        self._harmony_memory = []
        self._best_solution = None
    
    def initialize_population(self):
        """Initialize the harmony memory"""
        self._harmony_memory = []
        for _ in range(self._hms):
            point = [random.uniform(self._min, self._max) for _ in range(self._ndims)]
            if self._repair_function:
                point = self._repair_function(point)
                
            fitness = self._objective_function(self._preprocess_function(point))
            if fitness is not False and fitness is not None:
                self._harmony_memory.append(SolutionBasic(point, fitness))
        
        if self._harmony_memory:
            self._best_solution = self.find_best_solution(self._harmony_memory)
    
    def _memory_consideration(self, dimension):
        """Select a value from harmony memory"""
        random_harmony = random.choice(self._harmony_memory)
        return random_harmony.point[dimension]
    
    def _pitch_adjustment(self, value):
        """Adjust the pitch of a note"""
        if random.random() < self._par:
            return value + self._bw * random.uniform(-1, 1)
        return value
    
    def _create_new_harmony(self):
        """Generate a new harmony"""
        new_harmony = []
        for i in range(self._ndims):
            if random.random() < self._hmcr:
                # Memory consideration
                value = self._memory_consideration(i)
                # Pitch adjustment
                value = self._pitch_adjustment(value)
            else:
                # Random selection
                value = random.uniform(self._min, self._max)
            new_harmony.append(value)
        
        # Apply boundary conditions
        return [max(self._min, min(self._max, val)) for val in new_harmony]
    
    def _update_harmony_memory(self, new_harmony, new_fitness):
        """Update harmony memory if new harmony is better than worst harmony"""
        worst_index = 0
        worst_fitness = self._harmony_memory[0].fitness
        
        for i in range(1, self._hms):
            if ((self._to_max and self._harmony_memory[i].fitness < worst_fitness) or
                (not self._to_max and self._harmony_memory[i].fitness > worst_fitness)):
                worst_index = i
                worst_fitness = self._harmony_memory[i].fitness
        
        if ((self._to_max and new_fitness > worst_fitness) or
            (not self._to_max and new_fitness < worst_fitness)):
            self._harmony_memory[worst_index] = SolutionBasic(new_harmony, new_fitness)
    
    def run_yielded(self, iterations: int = 100, population: int = 30, hmcr: float = 0.9, par: float = 0.3, bw: float = 0.2, seed: int = None, verbose: bool = False):
        """Execute the Harmony Search algorithm with yielding"""
        self._iterations = iterations
        self._hms = population  # Use population as harmony memory size
        self._hmcr = hmcr  # Update harmony memory considering rate
        self._par = par  # Update pitch adjustment rate
        self._bw = bw  # Update bandwidth
        self._seed = seed
        
        random.seed(seed)
        np.random.seed(seed)
        
        self.initialize_population()
        
        if not self._harmony_memory:
            return
            
        self._best_solution = self.find_best_solution(self._harmony_memory)
        if not self._best_solution:
            return
            
        iteration = 1
        best_solution_historical = self._best_solution
        best_fitness_historical = best_solution_historical.fitness
        best_point_historical = np.copy(best_solution_historical.point)
        
        # yield initial state
        points = [e.point for e in self._harmony_memory]
        fts = [e.fitness for e in self._harmony_memory]
        bin_point = self._preprocess_function(best_point_historical)
        yield iteration, best_fitness_historical, bin_point, points, fts
            
        for it in range(iterations):
            if verbose:
                print("it: ", it+1, " fitness mejor: ", best_fitness_historical)
                
            # Create new harmony
            new_harmony = self._create_new_harmony()
            
            if self._repair_function:
                new_harmony = self._repair_function(new_harmony)
                
            # Evaluate new harmony
            new_fitness = self._objective_function(self._preprocess_function(new_harmony))
            
            if new_fitness is not False and new_fitness is not None:
                # Update harmony memory
                self._update_harmony_memory(new_harmony, new_fitness)
                
                # Update best solution
                if self._harmony_memory:
                    current_best = self.find_best_solution(self._harmony_memory)
                    if current_best and self._best_solution:
                        if ((self._to_max and current_best.fitness > self._best_solution.fitness) or
                            (not self._to_max and current_best.fitness < self._best_solution.fitness)):
                            self._best_solution = current_best
            
            iteration += 1
            
            # Update best solution
            if ((self._to_max and self._best_solution.fitness > best_fitness_historical) or
                (not self._to_max and self._best_solution.fitness < best_fitness_historical)):
                best_fitness_historical = self._best_solution.fitness
                best_point_historical = np.copy(self._best_solution.point)
                
            # yield current state
            points = [e.point for e in self._harmony_memory]
            fts = [e.fitness for e in self._harmony_memory]
            bin_point = self._preprocess_function(best_point_historical)
            yield iteration, best_fitness_historical, bin_point, points, fts
    
    def run(self, iterations: int = 100, population: int = 30, hmcr: float = 0.9, par: float = 0.3, bw: float = 0.2, seed: int = None, verbose: bool = False):
        """Execute the Harmony Search algorithm"""
        for _, best_fitness_historical, bin_point, _, _ in self.run_yielded(
            iterations, population, hmcr, par, bw, seed, verbose):
            continue
        return best_fitness_historical, bin_point



import random
import numpy as np
from ..IMetaheuristic import IMetaheuristic
from ..ISolution import SolutionBasic

class CuckooSearch(IMetaheuristic):
    def __init__(self, min_value, max_value, ndims, to_max=True, objective_function=None, repair_function=None, preprocess_function=None, population_size=25, pa=0.25):
        """
        Initialize Cuckoo Search Algorithm
        
        Args:
            min_value: Minimum value for variables
            max_value: Maximum value for variables
            ndims: Number of dimensions
            to_max: Whether to maximize (True) or minimize (False)
            objective_function: Function to optimize
            repair_function: Function to repair invalid solutions
            preprocess_function: Function to preprocess solutions
            population_size: Number of nests
            pa: Probability of egg abandonment
        """
        super().__init__(to_max)
        self._min = min_value
        self._max = max_value
        self._ndims = ndims
        self._objective_function = objective_function
        self._repair_function = repair_function if repair_function is not None else lambda p: p
        self._preprocess_function = preprocess_function if preprocess_function is not None else lambda p: p
        self._population_size = population_size
        self._pa = pa  # probability of alien eggs discovered
        self._nests = []
        self._best_solution = None
        
    def initialize_population(self):
        """Initialize the population of nests"""
        self._nests = []
        for _ in range(self._population_size):
            point = [random.uniform(self._min, self._max) for _ in range(self._ndims)]
            if self._repair_function:
                point = self._repair_function(point)
            if self._preprocess_function:
                point = self._preprocess_function(point)
                
            fitness = self._objective_function(point)
            if fitness is not False and fitness is not None:
                self._nests.append(SolutionBasic(point, fitness))
        
        # Ensure we have at least one valid solution
        while len(self._nests) < 1:
            point = [random.uniform(self._min, self._max) for _ in range(self._ndims)]
            if self._repair_function:
                point = self._repair_function(point)
            if self._preprocess_function:
                point = self._preprocess_function(point)
                
            fitness = self._objective_function(point)
            if fitness is not False and fitness is not None:
                self._nests.append(SolutionBasic(point, fitness))
        
        if self._nests:
            self._best_solution = self.find_best_solution(self._nests)
    
    def _levy_flight(self):
        """Generate steps using Levy Flight"""
        import math
        beta = 3/2
        sigma = (math.gamma(1 + beta) * math.sin(math.pi * beta / 2) / 
                (math.gamma((1 + beta) / 2) * beta * 2**((beta - 1) / 2)))**(1 / beta)
        u = np.random.normal(0, sigma, self._ndims)
        v = np.random.normal(0, 1, self._ndims)
        step = u / abs(v)**(1 / beta)
        return step
    
    def _get_cuckoo(self, current_nest):
        """Generate a new solution via Levy flight"""
        step_size = 0.01  # can be adjusted
        step = self._levy_flight()
        new_position = []
        
        for i in range(self._ndims):
            new_pos = current_nest.point[i] + step_size * step[i]
            new_position.append(new_pos)
            
        # Apply boundary conditions
        new_position = [max(self._min, min(self._max, pos)) for pos in new_position]
        return new_position
    
    def _abandon_nests(self):
        """Abandon worse nests and build new ones"""
        for i in range(len(self._nests)):
            if random.random() < self._pa:
                # Generate new nest
                new_point = [random.uniform(self._min, self._max) for _ in range(self._ndims)]
                if self._repair_function:
                    new_point = self._repair_function(new_point)
                if self._preprocess_function:
                    new_point = self._preprocess_function(new_point)
                    
                fitness = self._objective_function(new_point)
                if fitness is not False and fitness is not None:
                    self._nests[i] = SolutionBasic(new_point, fitness)
    
    def run_yielded(self, iterations: int = 100, population: int = 30, pa: float = 0.25, seed: int = None, verbose: bool = False):
        """Execute the Cuckoo Search algorithm with yielding"""
        self._iterations = iterations
        self._population_size = population  # Update population size
        self._pa = pa  # Update pa parameter
        self._seed = seed
        
        random.seed(seed)
        np.random.seed(seed)
        
        self.initialize_population()
        
        if not self._nests or not self._best_solution:
            return
            
        iteration = 1
        best_solution_historical = self._best_solution
        best_fitness_historical = best_solution_historical.fitness
        best_point_historical = np.copy(best_solution_historical.point)
        
        # yield initial state
        points = [e.point for e in self._nests]
        fts = [e.fitness for e in self._nests]
        bin_point = self._preprocess_function(best_point_historical)
        yield iteration, best_fitness_historical, bin_point, points, fts
            
        for it in range(iterations):
            if verbose:
                print("it: ", it+1, " fitness mejor: ", best_fitness_historical)
                
            # Get a random nest
            i = random.randint(0, len(self._nests) - 1)
            cuckoo_nest = self._nests[i]
            
            # Generate new solution via Levy flight
            new_position = self._get_cuckoo(cuckoo_nest)
            
            if self._repair_function:
                new_position = self._repair_function(new_position)
            if self._preprocess_function:
                new_position = self._preprocess_function(new_position)
                
            new_fitness = self._objective_function(new_position)
            
            if new_fitness is not False and new_fitness is not None:
                # Random nest to compare with
                j = random.randint(0, len(self._nests) - 1)
                
                # Replace if better
                if ((self._to_max and new_fitness > self._nests[j].fitness) or
                    (not self._to_max and new_fitness < self._nests[j].fitness)):
                    self._nests[j] = SolutionBasic(new_position, new_fitness)
            
            # Abandon worst nests and generate new ones
            self._abandon_nests()
            
            # Update best solution
            if self._nests:
                current_best = self.find_best_solution(self._nests)
                if current_best and self._best_solution:
                    if ((self._to_max and current_best.fitness > self._best_solution.fitness) or
                        (not self._to_max and current_best.fitness < self._best_solution.fitness)):
                        self._best_solution = current_best
        
            iteration += 1
            
            # Update best solution
            if ((self._to_max and self._best_solution.fitness > best_fitness_historical) or
                (not self._to_max and self._best_solution.fitness < best_fitness_historical)):
                best_fitness_historical = self._best_solution.fitness
                best_point_historical = np.copy(self._best_solution.point)
                
            # yield current state
            points = [e.point for e in self._nests]
            fts = [e.fitness for e in self._nests]
            bin_point = self._preprocess_function(best_point_historical)
            yield iteration, best_fitness_historical, bin_point, points, fts
    
    def run(self, iterations: int = 100, population: int = 30, pa: float = 0.25, seed: int = None, verbose: bool = False):
        """Execute the Cuckoo Search algorithm"""
        for _, best_fitness_historical, bin_point, _, _ in self.run_yielded(
            iterations, population, pa, seed, verbose):
            continue
        return best_fitness_historical, bin_point



#functions taken from https://doi.org/10.1007/s00521-017-3088-3
#Applying modified cuckoo search algorithm for solving systems of nonlinear equations
import numpy as np

class F1:
    @staticmethod
    def func_simple(x: float) -> float:
        result = 2 * (x - 0.75)**2
        result = result + np.sin(5*np.pi*x - 0.4*np.pi)
        result = result - 0.125
        return result
    
    @staticmethod
    def func(x_array: list) -> float:
        return F1.func_simple(x_array[0])

    @staticmethod
    def get_limits():
        return [0, 1]
    
    @staticmethod
    def get_theoretical_optimum():
        return -1.12323
    
    @staticmethod
    def get_type():
        return "min"

class F3:
    @staticmethod
    def func_simple(x: float) -> float:
        result = 0
        for j in range(1, 5+1): #j = 1, 2, 3, 4, 5
            result += j * np.sin( (j+1)*x +j )
        return -1*result

    @staticmethod
    def func(x_array: list) -> float:
        return F3.func_simple(x_array[0])

    @staticmethod
    def get_limits():
        return [-10, 10]
    
    @staticmethod
    def get_theoretical_optimum():
        return -12.03125
    
    @staticmethod
    def get_type():
        return "min"



class SolutionBasic:
    def __init__(self, point, fitness):
        self.point = point
        self.fitness = fitness

    def move_to(self, new_point, new_fitness):
        self.point = new_point
        self.fitness = new_fitness
        return self.point
    

class SolutionWithId(SolutionBasic):
    def __init__(self, _id, point, fitness):
        self.point = point
        self._id = _id
        self.fitness = fitness
    
    def set_id(self, _id):
        self._id = _id
    def get_id(self):
        return self._id



def standard(x: float, uniform_value: float):
    return 1 if uniform_value <= x else 0

def complement(x: float, uniform_value1: float, uniform_value2: float):
    return standard(1 - x, uniform_value2) if uniform_value1 <= x else 0

def static_probability(x: float, alpha: float, uniform_value: float):
    if alpha >= x:
        return 0
    else:
        if (alpha < x and x <= ((1 + alpha) / 2)):
            return standard(x, uniform_value)
        else:
            return 1

def elitist(x: float, uniform_value1: float, uniform_value2: float):
    return standard(x, uniform_value2) if uniform_value1 < x else 0



import random
import numpy as np
from ..IMetaheuristic import IMetaheuristic
from ..ISolution import SolutionBasic
from typing import List, Callable

class ArtificialBeeColony(IMetaheuristic):
    def __init__(self, min_value: float, max_value: float, ndims: int, to_max: bool,
     objective_function: Callable[[List[float]], float],
      repair_function: Callable[[List[float]], List[float]],
      preprocess_function: Callable[[List[float]], List[float]] = None):
        """
        Initialize Artificial Bee Colony Algorithm
        
        Args:
            to_max: Whether to maximize (True) or minimize (False)
            colony_size: Size of the colony (number of employed bees = number of onlooker bees)
            limit: Maximum number of trials before abandoning a food source
        """
        super().__init__(to_max)
        self._min = min_value
        self._max = max_value
        self._ndims = ndims
        self._to_max = to_max

        self.objective_function = objective_function
        self.preprocess_function = \
         preprocess_function if preprocess_function is not None else lambda p: p
        self.repair_function = \
         repair_function if repair_function is not None else lambda p: p

        self._food_sources = []  # food sources / solutions
        self._trials = []  # trial counter for each food source
        self._best_solution = None
    
    def initialize_population(self, population_size):
        """Initialize food sources and their trial counters"""
        self._food_sources = []
        self._trials = []
        
        for _ in range(population_size):
            point = [random.uniform(self._min, self._max) for _ in range(self._ndims)]
            point = self.cut_mod_point(point, self._min, self._max)
            # Use repair function if point is invalid
            repaired_point = self.repair_function(point)
            fitness = self.objective_function(self.preprocess_function(repaired_point))
            if fitness is not False and fitness is not None:
                self._food_sources.append(SolutionBasic(repaired_point, fitness))
                self._trials.append(0)
        
        # Ensure we have at least one valid solution
        while len(self._food_sources) < 1:
            point = [random.uniform(self._min, self._max) for _ in range(self._ndims)]
            point = self.cut_mod_point(point, self._min, self._max)
            repaired_point = self.repair_function(point)
            fitness = self.objective_function(self.preprocess_function(repaired_point))
            if fitness is not False and fitness is not None:
                self._food_sources.append(SolutionBasic(repaired_point, fitness))
                self._trials.append(0)
            
        self._best_solution = self.find_best_solution(self._food_sources)
    
    def _calculate_probability(self, solution):
        """Calculate probability for onlooker bee selection"""
        if self._to_max:
            return 0.9 * solution.fitness / self._best_solution.fitness + 0.1
        else:
            return 1.0 / (1.0 + abs(solution.fitness))
    
    def _generate_new_position(self, current, partner):
        """Generate new food source position"""
        new_position = []
        phi = random.uniform(-1, 1)
        
        for i in range(self._ndims):
            # Generate new position component
            new_pos = (current.point[i] + 
                      phi * (current.point[i] - partner.point[i]))
            new_position.append(new_pos)
        
        return self.cut_mod_point(new_position, self._min, self._max)
    
    def _employed_bee_phase(self):
        """Employed bee phase"""
        for i in range(self._colony_size):
            # Select random partner, different from current bee
            partner_idx = i
            while partner_idx == i:
                partner_idx = random.randint(0, self._colony_size - 1)
            
            # Generate new food source position
            new_position = self._generate_new_position(
                self._food_sources[i], 
                self._food_sources[partner_idx]
            )
            
            # Evaluate new position
            new_position = self.cut_mod_point(new_position, self._min, self._max)
            repaired_position = self.repair_function(new_position)
            new_fitness = self.objective_function(self.preprocess_function(repaired_position))
            
            # Replace if better and fitness is valid
            if (new_fitness is not False and new_fitness is not None and
                ((self._to_max and new_fitness > self._food_sources[i].fitness) or
                (not self._to_max and new_fitness < self._food_sources[i].fitness))):
                self._food_sources[i] = SolutionBasic(repaired_position, new_fitness)
                self._trials[i] = 0
            else:
                self._trials[i] += 1
    
    def _onlooker_bee_phase(self):
        """Onlooker bee phase"""
        # Calculate selection probabilities
        probabilities = [self._calculate_probability(source) for source in self._food_sources]
        prob_sum = sum(probabilities)
        probabilities = [p/prob_sum for p in probabilities]
        
        # For each onlooker bee
        for _ in range(self._colony_size):
            # Select food source based on probability
            selected_idx = np.random.choice(self._colony_size, p=probabilities)
            
            # Select random partner
            partner_idx = selected_idx
            while partner_idx == selected_idx:
                partner_idx = random.randint(0, self._colony_size - 1)
            
            # Generate new food source position
            new_position = self._generate_new_position(
                self._food_sources[selected_idx], 
                self._food_sources[partner_idx]
            )
            
            # Evaluate new position
            new_position = self.cut_mod_point(new_position, self._min, self._max)
            repaired_position = self.repair_function(new_position)
            new_fitness = self.objective_function(self.preprocess_function(repaired_position))
            
            # Replace if better and fitness is valid
            if (new_fitness is not False and new_fitness is not None and
                ((self._to_max and new_fitness > self._food_sources[selected_idx].fitness) or
                (not self._to_max and new_fitness < self._food_sources[selected_idx].fitness))):
                self._food_sources[selected_idx] = SolutionBasic(repaired_position, new_fitness)
                self._trials[selected_idx] = 0
            else:
                self._trials[selected_idx] += 1
    
    def _scout_bee_phase(self):
        """Scout bee phase"""
        for i in range(self._colony_size):
            # If trials limit exceeded, replace with new random solution
            if self._trials[i] >= self._limit:
                point = [random.uniform(self._min, self._max) for _ in range(self._ndims)]
                point = self.cut_mod_point(point, self._min, self._max)
                repaired_point = self.repair_function(point)
                fitness = self.objective_function(self.preprocess_function(repaired_point))
                if fitness is not False and fitness is not None:
                    self._food_sources[i] = SolutionBasic(repaired_point, fitness)
                    self._trials[i] = 0
    
    def run_yielded(self, iterations: int = 100, population: int = 30, limit: int = 20, seed: int = None, verbose: bool = False):
        """Execute the Artificial Bee Colony algorithm with yielding"""
        self._iterations = iterations
        self._population = population
        self._seed = seed
        self._colony_size = population // 2  # number of employed bees = number of food sources
        self._limit = limit  # limit of trials for abandonment
        
        random.seed(seed)
        np.random.seed(seed)

        self.initialize_population(self._colony_size)
        
        iteration = 1
        best_solution_historical = self.find_best_solution(self._food_sources)
        best_fitness_historical = best_solution_historical.fitness
        best_point_historical = np.copy(best_solution_historical.point)
        
        # yield initial state
        points = [e.point for e in self._food_sources]
        fts = [e.fitness for e in self._food_sources]
        bin_point = self.preprocess_function(best_point_historical)
        yield iteration, best_fitness_historical, bin_point, points, fts
        
        while iteration <= iterations:
            if verbose:
                print("it: ", iteration, " fitness mejor: ", best_fitness_historical)
                
            # Employed bee phase
            self._employed_bee_phase()
            
            # Onlooker bee phase
            self._onlooker_bee_phase()
            
            # Scout bee phase
            self._scout_bee_phase()
            
            iteration += 1
            
            # Update best solution
            current_best = self.find_best_solution(self._food_sources)
            if ((self._to_max and current_best.fitness > best_fitness_historical) or
                (not self._to_max and current_best.fitness < best_fitness_historical)):
                best_fitness_historical = current_best.fitness
                best_point_historical = np.copy(current_best.point)
                
            # yield current state
            points = [e.point for e in self._food_sources]
            fts = [e.fitness for e in self._food_sources]
            bin_point = self.preprocess_function(best_point_historical)
            yield iteration, best_fitness_historical, bin_point, points, fts
        
        # Final yield
        bin_point = self.preprocess_function(best_point_historical)
        points = [e.point for e in self._food_sources]
        fts = [e.fitness for e in self._food_sources]
        yield iteration, best_fitness_historical, bin_point, points, fts

    def run(self, iterations: int = 100, population: int = 30, limit: int = 20, seed: int = None, verbose: bool = False):
        """Execute the Artificial Bee Colony algorithm"""
        for _, best_fitness_historical, bin_point, _, _ in self.run_yielded(
            iterations, population, limit, seed, verbose):
            continue
        return best_fitness_historical, bin_point



import random
import numpy as np
from ..IMetaheuristic import IMetaheuristic
from ..ISolution import SolutionBasic
from typing import List, Callable

class AntColony(IMetaheuristic):
    def __init__(self, min_value: float, max_value: float, ndims: int, to_max: bool,
     objective_function: Callable[[List[float]], float],
      repair_function: Callable[[List[float]], List[float]],
      preprocess_function: Callable[[List[float]], List[float]] = None):
        """
        Initialize Ant Colony Optimization
        
        Args:
            to_max: Whether to maximize (True) or minimize (False)
            n_ants: Number of ants in the colony
            evaporation_rate: Rate of pheromone evaporation
            alpha: Pheromone importance factor
            beta: Heuristic information importance factor
        """
        super().__init__(to_max)
        self._min = min_value
        self._max = max_value
        self._ndims = ndims
        self._to_max = to_max

        self.objective_function = objective_function
        self.preprocess_function = \
         preprocess_function if preprocess_function is not None else lambda p: p
        self.repair_function = \
         repair_function if repair_function is not None else lambda p: p

        self._ants = []
        self._pheromone = None
        self._best_solution = None
        self._discretization_points = 100  # Number of points for discretization
        
    def initialize_population(self, population_size):
        """Initialize the ant colony and pheromone matrix"""
        self._ants = []
        # Initialize ants with random positions
        for _ in range(population_size):
            point = [random.uniform(self._min, self._max) for _ in range(self._ndims)]
            point = self.cut_mod_point(point, self._min, self._max)
            repaired_point = self.repair_function(point)
            fitness = self.objective_function(self.preprocess_function(repaired_point))
            if fitness is not False and fitness is not None:
                self._ants.append(SolutionBasic(repaired_point, fitness))
        
        # Ensure we have at least one valid solution
        while len(self._ants) < 1:
            point = [random.uniform(self._min, self._max) for _ in range(self._ndims)]
            point = self.cut_mod_point(point, self._min, self._max)
            repaired_point = self.repair_function(point)
            fitness = self.objective_function(self.preprocess_function(repaired_point))
            if fitness is not False and fitness is not None:
                self._ants.append(SolutionBasic(repaired_point, fitness))
        
        # Initialize pheromone matrix for each dimension
        # We discretize the continuous space into a grid
        self._pheromone = np.ones((self._ndims, self._discretization_points)) * 0.1
        
        self._best_solution = self.find_best_solution(self._ants)
    
    def _discretize_position(self, position, dim):
        """Convert continuous position to discrete index"""
        normalized = (position - self._min) / (self._max - self._min)
        index = int(normalized * (self._discretization_points - 1))
        return max(0, min(index, self._discretization_points - 1))
    
    def _continuous_position(self, index, dim):
        """Convert discrete index to continuous position"""
        normalized = index / (self._discretization_points - 1)
        return self._min + normalized * (self._max - self._min)
    
    def _construct_solution(self):
        """Construct a new solution for an ant"""
        solution = []
        for dim in range(self._ndims):
            # Calculate probabilities for each discretized point
            pheromone = self._pheromone[dim]
            # Simple heuristic information (can be modified based on problem)
            heuristic = np.ones(self._discretization_points)
            
            probabilities = (pheromone ** self._alpha) * (heuristic ** self._beta)
            probabilities = probabilities / np.sum(probabilities)
            
            # Choose position based on probabilities
            chosen_index = np.random.choice(self._discretization_points, p=probabilities)
            position = self._continuous_position(chosen_index, dim)
            solution.append(position)
        
        return solution
    
    def _update_pheromones(self):
        """Update pheromone levels"""
        # Evaporation
        self._pheromone *= (1 - self._evaporation_rate)
        
        # Add new pheromones based on solutions
        for ant in self._ants:
            deposit = 1.0 / (1.0 + abs(ant.fitness))  # Convert fitness to positive value
            for dim in range(self._ndims):
                idx = self._discretize_position(ant.point[dim], dim)
                self._pheromone[dim][idx] += deposit
        
        # Add extra pheromone for best solution
        best_deposit = 1.0 / (1.0 + abs(self._best_solution.fitness))
        for dim in range(self._ndims):
            idx = self._discretize_position(self._best_solution.point[dim], dim)
            self._pheromone[dim][idx] += best_deposit
    
    def run_yielded(self, iterations: int = 100, population: int = 30, evaporation_rate: float = 0.1, alpha: float = 1.0, beta: float = 2.0, seed: int = None, verbose: bool = False):
        """Execute the Ant Colony Optimization algorithm with yielding"""
        self._iterations = iterations
        self._population = population
        self._seed = seed
        self._n_ants = population
        self._evaporation_rate = evaporation_rate
        self._alpha = alpha
        self._beta = beta
        
        random.seed(seed)
        np.random.seed(seed)

        self.initialize_population(self._n_ants)
        
        iteration = 1
        best_solution_historical = self.find_best_solution(self._ants)
        best_fitness_historical = best_solution_historical.fitness
        best_point_historical = np.copy(best_solution_historical.point)
        
        # yield initial state
        points = [e.point for e in self._ants]
        fts = [e.fitness for e in self._ants]
        bin_point = self.preprocess_function(best_point_historical)
        yield iteration, best_fitness_historical, bin_point, points, fts
        
        while iteration <= iterations:
            if verbose:
                print("it: ", iteration, " fitness mejor: ", best_fitness_historical)
                
            # Generate new solutions for all ants
            new_ants = []
            for _ in range(self._n_ants):
                new_position = self._construct_solution()
                new_position = self.cut_mod_point(new_position, self._min, self._max)
                repaired_position = self.repair_function(new_position)
                fitness = self.objective_function(self.preprocess_function(repaired_position))
                if fitness is not False and fitness is not None:
                    new_ants.append(SolutionBasic(repaired_position, fitness))
            
            if new_ants:  # Only update if we have valid ants
                self._ants = new_ants
            
            # Update best solution
            current_best = self.find_best_solution(self._ants)
            if ((self._to_max and current_best.fitness > best_fitness_historical) or
                (not self._to_max and current_best.fitness < best_fitness_historical)):
                best_fitness_historical = current_best.fitness
                best_point_historical = np.copy(current_best.point)
                self._best_solution = current_best
            
            # Update pheromone trails
            self._update_pheromones()
            
            iteration += 1
                
            # yield current state
            points = [e.point for e in self._ants]
            fts = [e.fitness for e in self._ants]
            bin_point = self.preprocess_function(best_point_historical)
            yield iteration, best_fitness_historical, bin_point, points, fts
        
        # Final yield
        bin_point = self.preprocess_function(best_point_historical)
        points = [e.point for e in self._ants]
        fts = [e.fitness for e in self._ants]
        yield iteration, best_fitness_historical, bin_point, points, fts

    def run(self, iterations: int = 100, population: int = 30, evaporation_rate: float = 0.1, alpha: float = 1.0, beta: float = 2.0, seed: int = None, verbose: bool = False):
        """Execute the Ant Colony Optimization algorithm"""
        for _, best_fitness_historical, bin_point, _, _ in self.run_yielded(
            iterations, population, evaporation_rate, alpha, beta, seed, verbose):
            continue
        return best_fitness_historical, bin_point



#This file implements the sudoku solving problem
import numpy as np
from anmetal.problems.IProblem import IProblem
class Sudoku(IProblem):
    def __init__(self, initial_state: list=None):
        self.state = initial_state
        if self.state is None:
            self.state = np.zeros((9,9))
        self.categories = [1,2,3,4,5,6,7,8,9]

    def set_categories(self, categories):
        self.categories = categories

    def is_row_violating(self, row, comming_from: str = ""):
        if len(row) != 9:
            raise ValueError("state row, column or box should be of size 9. comming from: "+comming_from)
        ncats = 0
        violated_row = False
        for x in row:
            if x != 0 and x not in self.categories:  # Allow 0 for empty cells
                violated_row = True
                break
            if x != 0:  # Only count non-empty cells
                equals_filter = row == x
                if np.sum(equals_filter) > 1:
                    violated_row = True
                    break
                ncats += 1
        return violated_row

    def get_violations(self):
        if not isinstance(self.state, np.ndarray):
            raise ValueError("state should be numpy ndarray")
        n_violations = 0
        #horizontal
        for ihoriz in range(9):
            if self.is_row_violating(self.state[ihoriz], "horizontal: "+str(ihoriz)):
                n_violations += 1
        #vertical
        for ivert in range(9):
            if self.is_row_violating(self.state[:,ivert], "vertical: "+str(ivert)):
                n_violations += 1
        #boxes
        for ih1 in [0,3,6]:
            ih2 = ih1+3
            for ivert1 in [0,3,6]:
                ivert2 = ivert1 + 3
                box = self.state[ih1:ih2,ivert1:ivert2]
                if self.is_row_violating(box.flatten(), "box ["+str(ih1)+":"+str(ih2)+","+str(ivert1)+":"+str(ivert2)+"]"):
                    n_violations +=1
        return n_violations

    def is_valid_solution(self, solution=None):
        """Check if the current state or provided solution is valid."""
        if solution is not None:
            old_state = self.state
            self.state = np.array(solution)
            is_valid = self.get_violations() == 0
            self.state = old_state
            return is_valid
        return self.get_violations() == 0



import math

def sShape1(x: float):
    return (1 / (1 + math.pow(math.e, -2 * x)))

def sShape2(x: float):
    return (1 / (1 + math.pow(math.e, -x)))

def sShape3(x: float):
    return (1 / (1 + math.pow(math.e, -x / 2)))

def sShape4(x: float):
    return (1 / (1 + math.pow(math.e, -x / 3)))

def vShape1(x: float):
    return abs(erf((math.sqrt(math.pi) / 2) * x))

def vShape2(x: float):
    return abs(math.tanh(x))

def vShape3(x: float):
    return abs(x / math.sqrt(1 + math.pow(x, 2)))

def vShape4(x: float):
    return abs((2 / math.pi) * math.atan((math.pi / 2) * x))

def erf(z: float):
    q: float = 1.0 / (1.0 + 0.5 * abs(z))
    ans: float = 1 - q * math.exp(-z * z - 1.26551223
            + q * (1.00002368
            + q * (0.37409196
            + q * (0.09678418
            + q * (-0.18628806
            + q * (0.27886807
            + q * (-1.13520398
            + q * (1.48851587
            + q * (-0.82215223
            + q * (0.17087277))))))))))
    return ans if z >= 0 else -ans



from anmetal.problems.IProblem import IProblem

import numpy as np
from numpy.random import RandomState
class Partition_Subset_Abstract(IProblem):
    def __init__(self, seed:int, num_dims:int, data=None):
        if data is not None and type(data) == type([]):
            self.data = data
        else:
            rnd = RandomState(seed)
            self.data = rnd.uniform(0, num_dims-1, num_dims)
            #mydata = rnd.randint(-100, 100, 200)
        self.ndim = len(self.data)+2
        self.min_x = 0
        self.max_x = len(self.data)


    def is_valid(self, point):
        are_1s = [0 for _ in range(len(self.data))]
        for i in range(2, len(point)):
            if point[i] < 0 or int(point[i]) >= len(self.data):
                return False
            are_1s[int(point[i])] += 1
            if are_1s[int(point[i])] > 1:
                return False
        for el in are_1s:
            if el == 0 or el > 1:
                return False
        return True

    def repair_function(self, point):
        point = np.abs(point)
        are_not = []
        are_1s = [0 for _ in range(len(self.data))]
        #dups = []
        dups = [[] for _ in range(len(self.data))]
        for i in range(2, len(point)):
            if int(point[i]) < 0 or int(point[i]) >= len(self.data):
                point[i] = 0
            are_1s[int(point[i])] += 1
            dups[int(point[i])].append(i)
        for i in range(len(are_1s)):
            if are_1s[i] == 0:
                are_not.append(i)
        sum_dups = []
        for i in range(len(dups)):
            sum_dups.append(len(dups[i]))
            if len(dups[i]) > 0:
                del dups[i][0]
        dups = [e for di in dups for e in di] #flatten
        if len(are_not) != len(dups) or are_1s != sum_dups:
            print("rayooooooooooos son distintos")
            print("arenot: ", len(are_not), ", dups: ", len(dups))
            print("are not: ", are_not)
            print("dups: ", dups)
            print("are1s: ", are_1s)
            point2 = [int(e) for e in point]
            print("point int: ", point2)
            print("point: ", point)
            raise Exception("son distintos :c")
        for i in range(len(dups)):
            point[dups[i]] = are_not[i]
        return point

    def preprocess_function(self, point):
        point2 = np.abs(np.copy(point))
        for i in range(2, len(point2)):
            point2[i] = int(point2[i])
        return point2

    def objective_function(self, point):
        raise NotImplementedError
    
    def get_sums(self, point):
        if not self.is_valid(point):
            return (False, False)
        s1 = point[0]
        s2 = point[1]
        s1 = int(s1/float(s1+s2)*len(self.data))
        s1 = s1 if s1 >= 0 and s1 < len(self.data) else 0
        s2 = len(self.data)-s1
        sum1 = 0
        sum2 = 0
        for i in range(2, s1+2, 1):
            sum1 += self.data[int(point[i])]
        for i in range(s1+2, len(point), 1):
            sum2 += self.data[int(point[i])]
        return (sum1, sum2)

    def print_difference_subsets(self, point):
        s1 = point[0]
        s2 = point[1]
        s1 = int(s1/float(s1+s2)*len(self.data))
        s1 = s1 if s1 >= 0 and s1 < len(self.data) else 0
        s2 = len(self.data)-s1
        sum1 = 0
        sum2 = 0
        sub1 = []
        sub2 = []
        for i in range(2, s1+2, 1):
            sum1 += self.data[int(point[i])]
            sub1.append(self.data[int(point[i])])
        for i in range(s1+2, len(point), 1):
            sum2 += self.data[int(point[i])]
            sub2.append(self.data[int(point[i])])
        print("sum1:", sum1)
        print("sum2:", sum2)

class Partition_Real(Partition_Subset_Abstract):
    #def __init__(self, seed:int, num_dims:int, data=None):
    #    super().__init__(seed, num_dims, data)

    def objective_function(self, point):
        sum1, sum2 = self.get_sums(point)
        if sum1 is False:
            return False
        return abs(sum1-sum2)


class Subset_Real(Partition_Subset_Abstract):
    #def __init__(self, seed:int, num_dims:int, data=None):
    #    super().__init__(seed, num_dims, data)

    def objective_function(self, point):
        sum1, _ = self.get_sums(point)
        if sum1 is False:
            return False
        return np.abs(sum1)



from abc import ABC, ABCMeta, abstractmethod


class IMetaheuristic(ABC):
    __metaclass__ = ABCMeta

    def __init__(self, to_max):
        self._iterations = None
        self._population = None
        self._to_max = to_max
        self.objective_function = None
    
    @abstractmethod
    def run(self):
        raise NotImplementedError

    @abstractmethod
    def initialize_population(self):
        raise NotImplementedError
    
    def is_point_valid(self, point):
        fitness = self.objective_function(point)
        if not fitness:
            return False
        else:
            return True

    def find_best_point(self, points: list) -> list: #all points are valid
        best = None
        best_fitness = None
        for point in points:
            fitness = self.objective_function(point)
            if best is None:
                best = point
            if best_fitness is None:
                best_fitness = fitness
            if self._to_max and fitness > best_fitness:
                best = point
                best_fitness = fitness
            if not self._to_max and fitness < best_fitness:
                best = point
                best_fitness = fitness
        return best
    
    def find_best_solution(self, solutions):
        best = None
        best_fitness = None
        for sol in solutions:
            if best is None:
                best = sol
            if best_fitness is None:
                best_fitness = sol.fitness
            if self._to_max and sol.fitness > best_fitness:
                best = sol
                best_fitness = sol.fitness
            if not self._to_max and sol.fitness < best_fitness:
                best = sol
                best_fitness = sol.fitness
        return best
    
    @staticmethod
    def cut_mod_point(point, min_x, max_x):
        newpoint = []
        ran = abs(max_x-min_x)
        for x in point:
            if x < min_x:
                dist = abs(x - min_x)
                newnum = max_x-(dist%ran)
                newpoint.append(newnum)
            elif x > max_x:
                dist = abs(max_x - x)
                newnum = min_x+(dist%ran)
                newpoint.append(newnum)
            else:
                newpoint.append(x)
        return newpoint



"""Tests for optimization algorithms."""
import pytest
import numpy as np
from anmetal.population.AFSA.AFSAMH_Real import AFSAMH_Real
from anmetal.population.PSO.PSOMH_Real import PSOMH_Real
from anmetal.problems.nonlinear_functions.two_inputs import Goldsteinprice

class TestAFSA:
    def test_afsa_initialization(self):
        """Test AFSA algorithm initialization."""
        problem = Goldsteinprice()
        afsa = AFSAMH_Real(
            min_x=problem.get_limits()[0],
            max_x=problem.get_limits()[1],
            dimension=2,
            to_maximize=False,
            objective_function=problem.func,
            repair_function=None,
            preprocess_function=None
        )
        assert afsa.dimension == 2
        assert not afsa.to_maximize

    def test_afsa_run(self):
        """Test AFSA algorithm execution."""
        problem = Goldsteinprice()
        afsa = AFSAMH_Real(
            min_x=problem.get_limits()[0],
            max_x=problem.get_limits()[1],
            dimension=2,
            to_maximize=False,
            objective_function=problem.func,
            repair_function=None,
            preprocess_function=None
        )
        fit, pt = afsa.run(
            verbose=False,
            iterations=10,
            population=20,
            visual_distance_percentage=0.5,
            velocity_percentage=0.5,
            seed=115
        )
        assert isinstance(fit, (int, float))
        assert len(pt) == 2

class TestPSO:
    def test_pso_initialization(self):
        """Test PSO algorithm initialization."""
        problem = Goldsteinprice()
        pso = PSOMH_Real(
            min_x=problem.get_limits()[0],
            max_x=problem.get_limits()[1],
            dimension=2,
            to_maximize=False,
            objective_function=problem.func,
            repair_function=None,
            preprocess_function=None
        )
        assert pso.dimension == 2
        assert not pso.to_maximize

    def test_pso_run(self):
        """Test PSO algorithm execution."""
        problem = Goldsteinprice()
        pso = PSOMH_Real(
            min_x=problem.get_limits()[0],
            max_x=problem.get_limits()[1],
            dimension=2,
            to_maximize=False,
            objective_function=problem.func,
            repair_function=None,
            preprocess_function=None
        )
        fit, pt = pso.run(
            verbose=False,
            iterations=10,
            population=20,
            omega=0.8,
            phi_g=0.5,
            phi_p=0.5,
            seed=115
        )
        assert isinstance(fit, (int, float))
        assert len(pt) == 2



from anmetal.population.ISolution import SolutionWithId


class SolutionParticle(SolutionWithId):
    def __init__(self, _id, point, fitness, velocity):
        self.point = point
        self._id = _id
        self.fitness = fitness
        self.velocity = velocity
        self.set_best_point(point, fitness)
    
    def set_best_point(self, newpoint, newfitness):
        self.best_point = [x for x in newpoint]
        self.best_fitness = newfitness
    
    def set_velocity(self, _velocity):
        self.velocity = _velocity
    def get_velocity(self):
        return self.velocity

    def set_id(self, _id):
        self._id = _id
    def get_id(self):
        return self._id